#! /usr/bin/env python

"""
# ---------------------------------------------------------------------------
#
# functions_network_medicine_schizo.py
# File with all the functions that I have used to calculate the distance
# and the proximity between diseases and drugs. Additionally, functions are included
# employees to obtain the disease module and study its
# significance prior to determining distance and proximity.
#
# Exploring Drug Repurposing Opportunities for Schizophrenia: A Network Medicine Approach 
#
# María Marín Tercero
# ----------------------------------------------------------------------------
"""

import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import networkx as nx
from tabulate import tabulate
from networkx.algorithms import bipartite
import random
from scipy.stats import norm
from itertools import combinations
import re
from itertools import product

    

# =================================================================================

def genes_dis(enf, file):
    """
    This function creates a list with the genes associated with the disease "enf" in the dis_gen file   
    
    """
    genes=[]
    for i, dis in enumerate(file["dis"]): 
        if dis == enf:
            gen = file["gen"][i]
            genes.append(gen) 
    return genes

# =================================================================================

def pro_gen_dict(gene_list, file):
    """
    This function creates a dictionary from the list of genes associated with the disease with:
     key: protein associated with each gene in the gen_pro file
     value: gene related to the key protein in the gen_pro file
    """
    result_dict = {}
    for i, gen in enumerate(file["gen"]): 
        # Looping through gen_pro, which relates genes and proteins.
        # I'm storing the position of the gene (i) and the gene id (gen).
        if gen in gene_list: 
            # Searching each gene in gen_pro within the corresponding gene list of each disease.
            prot = file["pro"][i] 
            # If that gene is in the gene list of each disease, I find the associated protein at the same position.
            result_dict[prot] = gen 
            # Adding to each disease's dictionary the protein as key and the related gene as value.
    return result_dict

# =================================================================================

def gen_pro_PPI(dict1, file):
    """
    From a dictionary with the relationships between proteins and genes associated with each of our diseases,
    this function retains the prot:gen relationship from the dictionary only if such prot appears in the PPI network of the pro_pro file.
    key: proteins appearing in the PPI network
    value: genes related to the key protein
    """
    result_dict = {}
    for prot in dict1.keys(): 
        # Iterating over all proteins in the general prot:gen dictionary.
        if prot in file["prA"].tolist() or prot in file["prB"].tolist(): 
            # Selecting proteins that appear in the PPI network.
            result_dict[prot] = dict1[prot] 
            # Adding to the PPI prot:gen dictionary only the prot:gen relationships for proteins that are in the PPI.
    return result_dict 

# =================================================================================

def SG(dic, PPI):
    """ 
    Input data: dictionary with proteins from the PPI (keys) and associated genes (values) for a disease, PPI network
    This function creates a subgraph only with the proteins from the PPI network associated with my disease as nodes.
    """

    # Creating a subgraph only with the proteins from the PPI network associated with my disease as nodes
    SG = nx.subgraph(PPI, dic.keys())
    return SG

# =================================================================================

def lcc(SG):
    """
    This function gives us the LCC of the proteins from the PPI network associated with a disease from a subgraph
    formed only with the proteins associated with the disease.
    """
    lcc = max(nx.connected_components(SG), key=len) 
    # Calculating the LCC (module comprising the largest number of proteins associated with a disease).
    
    # Our goal is to obtain the number of genes that are part of the LCC of the disease:
    # The number of proteins from the disease in the LCC is the same number as the genes in the LCC
    # (because we have extracted the list of proteins from the dictionary where they form a tuple with their associated genes).
    
    return lcc

# =================================================================================

def nodes_by_degree(G):
    """
    This function returns a dictionary where we will obtain the degrees as keys and, in the values, all the nodes of the network that contain that degree.
    """
    degree_dict = {}
    for node in G.nodes():
        degree = G.degree(node)

        if degree not in degree_dict:
            degree_dict[degree] = []

        degree_dict[degree].append(node)

    return degree_dict

# =================================================================================

def lcc_simulation(G, lcc, PPI, dp=False):
    """
    Input data: disease module subgraph (G), PPI, and disease LCC (lcc). 
    Dp can be False to compute networks preserving the degree distribution of the disease module or True otherwise.
    
    This function returns the mean and standard deviation of the LCC of 1000 random networks with the same number 
    of nodes and edges as the graph G provided as input. 
    
    For the creation of these networks in the case dp = False, the associations between disease proteins are randomly 
    distributed within the network. Therefore, the networks will not have the same structure as the disease module.
    
    For the creation of networks in the case dp = True, nodes in the PPI network with the same degree as the nodes 
    in the disease module are selected. Thus, networks with the same structure as the disease module are created.
    """
    
    # Preliminary calculations:
    
    # Get the total proteins from the PPI
    ppi_nodes = PPI.nodes()
    
    # Get the total proteins in the disease module
    disease_nodes = G.nodes()
    
    # Get the number of proteins in the disease module
    num_disease_nodes = len(disease_nodes & set(ppi_nodes))
    
    # Group nodes by their degree
    degree_grouped_nodes = nodes_by_degree(PPI)


    # 1000 random simulations to calculate 1000 LCC
    random_list = []
    for i in range(1000):
        if dp == False: # Simulation for non-degree preserving
            # Get a random set of proteins within the total nodes of the PPI network.
            # This set has a number of nodes equivalent to the number of nodes in the disease module.
            random_nodes = set(random.sample(list(ppi_nodes), num_disease_nodes))
        if dp == True: # Simulation for degree preserving
            random_nodes = set()
            for node in disease_nodes: # For each node in the disease module
                degree = PPI.degree(node) # Get its degree in the PPI network
                available_nodes = degree_grouped_nodes[degree] # Get a group of nodes in the PPI with the same degree as the iterated node
                control = True
                while(control): # Loop to choose sampled nodes only once
                    chosen_node = random.choice(available_nodes) # Choose a node from the selected PPI nodes in the previous step
                    control = chosen_node in random_nodes # Check if that node is among the sampled nodes
                random_nodes.add(chosen_node) # Add that node to the list of nodes that I will use to create the random network
                # Only nodes that are not previously among the sampled nodes are added thanks to the while loop  
        r = nx.subgraph(PPI,random_nodes) # Subgraph of the PPI with the selected random nodes
        r = nx.Graph(r) # To remove parallel edges
        r.remove_edges_from(nx.selfloop_edges(r)) # To remove self-loops connecting a node to itself
        random_list.append(len(max(nx.connected_components(r), key=len)))
    mean = np.mean(random_list) # Mean
    std = np.std(random_list) # Standard deviation
    zscore = (len(lcc) - mean)/std
        
    return mean, std, zscore, random_list

# =================================================================================

def degrees_list(G):
    """
    This function returns a list with the nodes and another list with their degrees.
    """
    nodes = list(G.nodes())
    degrees = list(dict(G.degree()).values())
    
    return nodes, degrees

# =================================================================================


def rep(disease_name, G, PPI, ndp_list, dp_list, LCC, mean_ndp, std_ndp, zscore_ndp, mean_dp, std_dp, zscore_dp):
    """
    Input data:
    1. Disease name
    2. SG: LCC of the proteins from the PPI network associated with a disease
    3. G: PPI network 
    4. ndp_list: list of LCCs (ndp)
    5. dp_list: list of LCCs (dp)
    6.LCC: Observed LCC of the disease
    7. mean_ndp: mean of the list of LCCs (ndp)
    8. std_ndp: standard deviation of the list of LCCs (ndp)
    9.zscore_ndp: z-score of the list of LCCs (ndp)
    10. mean_dp: mean of the list of LCCs (dp) 
    11. std_dp: standard deviation of the list of LCCs (dp)
    12. zscore_dp: z-score of the list of LCCs (dp)
    
    This function performs a representation of:
    1. Distribution of LCCs (ndp)
    2. Distribution of LCCs (dp)
    """
    
    # Get a list with the nodes and their degrees from the PPI
    G_ppi_list = degrees_list(PPI)
    
    degree_ppi = pd.DataFrame(list(zip(G_ppi_list[0], G_ppi_list[1])), columns=['node','degree'])
     
    # Get a list with the nodes and their degrees from the PPI
    G_disease_list = degrees_list(G)
    
    degree_disease = pd.DataFrame(list(zip(G_disease_list[0], G_disease_list[1])), columns=['node','degree'])

      
    # Group the list of nodes and degrees from the PPI network by degree and count how many nodes in the network have that degree    
    G_plot_ppi = degree_ppi.groupby('degree').count()
    
    # Group the list of nodes and degrees from the disease by degree and count how many nodes in the network have that degree
    G_plot_disease = degree_disease.groupby('degree').count()
    
    # Create a figure to add 3 subplots
    fig, axs = plt.subplots(1, 2, figsize=(15, 4))

    # Representation of the distribution of LCC ndp
    axs[0].hist(ndp_list, color = '#79C4FF', bins=20, edgecolor='black')
    axs[0].set_xlabel('Nº of nodes in LCC', fontsize=12)
    axs[0].set_ylabel('Nº of random networks', fontsize=12)
    axs[0].set_title('LCC distribution for 1,000 random networks (ndp)', fontsize=14)
    axs[0].axvline(x=len(LCC), color='#FF7A7A', linestyle='--')
    axs[0].legend(["LCC "+str(disease_name)], loc='upper right')

    
    # Representation of the distribution of LCC dp
    axs[1].hist(dp_list, color = '#79C4FF', bins=20, edgecolor='black')
    axs[1].set_xlabel('Nº of nodes in LCC', fontsize=12)
    axs[1].set_ylabel('Nº of random networks', fontsize=12)
    axs[1].set_title('LCC distribution for 1,000 random networks (dp)', fontsize=14)
    axs[1].axvline(x=len(LCC), color='#FF7A7A', linestyle='--')
    axs[1].legend(["LCC "+str(disease_name)], loc='upper right')

    
    # Adjust the layout and space between plots
    plt.tight_layout()
    
    # Legend 1
    legend_text_1 = r'$LCC (ndp)_{\mathrm{obs}}$' + ' (' + r'$\mathrm{mean\ LCC} - \mathrm{STD}$, z-score' + ')'
    leyenda_valor_1 = f'{len(LCC)} ({round(mean_ndp,2)} - {round(std_ndp,2)}, {round(zscore_ndp,2)})'
    fig.text(0.51, -0.05, legend_text_1+str(" = ")+leyenda_valor_1, fontsize=12, ha='center', va='center', bbox=dict(facecolor='white', alpha=0.5, boxstyle='round'))
    
    # Legend 2 
    legend_text_2 = r'$LCC (dp)_{\mathrm{obs}}$' + ' (' + r'$\mathrm{mean\ LCC} - \mathrm{STD}$, z-score' + ')'
    leyenda_valor_2 = f'{len(LCC)} ({round(mean_dp,2)} - {round(std_dp,2)}, {round(zscore_dp,2)})'
    fig.text(0.51, -0.15, legend_text_2+str(" = ")+leyenda_valor_2, fontsize=12, ha='center', va='center', bbox=dict(facecolor='white', alpha=0.5, boxstyle='round'))
    
    # General title above the two plots
    plt.suptitle(str(disease_name), fontsize=20, y=1.1, fontweight='bold', style='italic')

    # Show the figure 
    plt.show()



# ================================================================================= 

def targets(drug_list, arch):
    """
    Input data: list of drugs treating a disease (drug_list) and a file with drug-target relationships (arch).
    This function allows us to obtain a DataFrame with drugs in the "Drugs" column and their targets in the "Targets" column.
    """
    targets_total = [] 
    # List to store targets for each drug separated by commas
    drugs = [] 
    # List to store drugs found in the 'arch' file
    for drug1 in drug_list: 
        # Iterate over each drug in the drug list
        targets = [] 
        # Empty list to store targets for each drug
        for i, drug2 in enumerate(arch["dru"]): 
            # Iterate over drugs in the 'arch' file, keeping track of the drug (drug2) and its row (i) in the drug column
            if drug1 == drug2: 
                # If a drug for a disease is found in the drug-target file
                targets.append(arch["pro"][i]) 
                # Add its target to the targets list, which will be in the same row (i) in the target column
        if len(targets) > 0: 
            # Check if the targets list is not empty
            drugs.append(drug1) 
            # Add the drug to the list of drugs, so only drugs appearing in the 'arch' file are stored
            targets_total.append(targets) 
            # Add the list of targets for that drug to the list of targets for all drugs
            
    data = {"Drugs": drugs, "Targets": targets_total} 
    # Combine the data and classify them into Drugs and Targets
    df = pd.DataFrame(data) 
    # Create a DataFrame with the results
    return df

# =================================================================================

def calculate_dc_drug(target_list, dist_matrix, disease_module_proteins, PPI):
    """
    Input data: list of targets of the drug (target_list), proteins of the disease module (disease_module_proteins), 
    matrix file with distances between all nodes of the PPI (dist_matrix), PPI network (PPI).
    This function returns the closest measure (dc) of a drug and a disease.
    """
    targets_in_ppi = set(target_list) & set(PPI.nodes()) 
    # List of targets of the drug that are also in the PPI network
    targets_in_disease_module = set(targets_in_ppi) & set(disease_module_proteins)
    # List of targets of the drug that are also part of the disease module
    
    distances_disease_target = dist_matrix.loc[list(targets_in_ppi), list(disease_module_proteins)].values  
    # Generate a matrix with the shortest path lengths (SPLs) between all drug targets and all proteins in the disease module according to the distance file
    
    non_empty_rows = ~np.isnan(distances_disease_target).all(axis=1) 
    # Keep rows that are not empty, i.e., remove targets that have no path to any protein in the disease module
    
    if np.isnan(distances_disease_target).all(): 
        # If the previous matrix is empty
        return np.nan 
        # There is no path between any drug target and the disease
    
    elif len(targets_in_disease_module) == len(targets_in_ppi): 
        # If all drug targets are part of the disease module
        return 0 
        # The dc value will be 0
    
    else: 
        return  np.nanmean(np.nanmin(distances_disease_target[non_empty_rows], axis=1)) 
        # Otherwise, calculate the mean of the minimum SPLs of targets that have a path to the disease module (the mean of the minimum values of each row of the matrix)

# ================================================================================= 

def proximity(df, arch_dist, prots_enf, PPI):
    """
    Input data: DataFrame (df) with "Drugs" and their "Targets"; file with distances between all nodes of the network (arch_dist);
    list of proteins related to a disease (prots_enf); and PPI network (PPI).
    Function that allows us to obtain a DataFrame with the proximity values (dc) for a list of drugs and a disease.
    """
    dc_total = []  # List to store the dc value of all drugs
    diseases = []  # List to add whether the drug belongs to a specific disease
    for i, drug in enumerate(df["Drugs"]):  # For each drug in the drugs and targets DataFrame, I keep the row (i)
        targets_list = df["Targets"][i]  # Get the list of targets for that drug, which will be in the same row as the drug but in the targets column
        dc_total.append(calculate_dc_drug(targets_list, arch_dist, prots_enf, PPI))

    data = {'Drugs': list(df["Drugs"]), 'dc': dc_total}  # Save the relationship between the list of drugs and the list of dcs
    result_table = pd.DataFrame(data)  # Convert the results into a DataFrame

    return result_table


# =================================================================================

def proximity_random(df, arch_dist, prots_enf, PPI, num_iterations, df_results):
    """
    Input data: DataFrame (df) with "Drugs" and their "Targets", file with distances between all nodes of the network (arch_dist),
    list of proteins related to a disease (prots_enf), PPI network (PPI), and the number of iterations (num_iterations)
    to calculate the dc (num_iterations), DataFrame (df_results) with the results of random target modules iterations for each drug.
    This function allows us to obtain a DataFrame with the average proximity values (dc) calculated from:
    - 1000 random modules of proteins with the same number of proteins and the same degree distribution as the disease module.
    - 1000 random target modules with the same number of proteins and the same degree distribution as each drug in the Drugs and Targets DataFrame.
    """
    group_nodes_degree = nodes_by_degree(PPI)  # Group nodes by their degree
    
    # Initialize a matrix with null values ​​that has the same number of rows as the total number of drugs in the DataFrame
    # and the same number of columns as the number of iterations
    proximity_matrix = np.full((len(df["Drugs"]), num_iterations), None, dtype=object)

    for i in range(num_iterations):  # For each iteration
        
        # Random disease module
        random_prots = set()  # Create a set of random proteins
        for prot in prots_enf:  # For each protein in the disease module
            degree_prot = PPI.degree(prot)  # Calculate its degree
            available_prots = group_nodes_degree[degree_prot]  # Choose proteins from the PPI with the same degree
            random_prots.add(np.random.choice(available_prots))  # Choose the same number of nodes as the disease module, taken from the total list of PPI proteins randomly

        # Random target module
        random_targets = {}  # Create a dictionary to store random targets for each drug
        for j, drug in enumerate(df["Drugs"]):  # For each drug in the DataFrame
            random_target = df_results.iloc[j, i]  # Get the random target corresponding to this iteration for this drug
            random_targets[drug] = random_target  # Store the random target in the dictionary
            
            # Calculate the dc of each drug with the random disease module and the random target
            drug_dc_total = calculate_dc_drug(random_target, arch_dist, random_prots, PPI)
            proximity_matrix[j, i] = drug_dc_total  # Add the dc of each drug (in row j) in the column of the matrix corresponding to the iteration (i)
            
    
    drug_mean_proximity = []  # List to add the mean of dc after 1000 iterations for each drug
    deviation = []  # List to add the deviation of dc after 1000 iterations for each drug
    
    for row in proximity_matrix:  # For each row (for each drug)
        if all(x is np.nan for x in row):  # If the entire row is None
            drug_mean_proximity.append(np.nan)  # Add None to the drug_mean_proximity list
            deviation.append(np.nan)  # Add None to the deviation list
        else:
            drug_mean_proximity.append(np.nanmean(row))  # Add the mean of that row
            deviation.append(np.nanstd(row))  # Standard deviation
            
    data = {'Drugs': list(df["Drugs"]), 'dc_mean': drug_mean_proximity, 'dc_std' : deviation}
    result_table = pd.DataFrame(data)

    return result_table



# ================================================================================= 

def calculate_random_drug_target_modules(num_iterations, df, PPI):
    results = {}

    group_nodes_by_degree = functions_network_medicine_schizo.nodes_by_degree(PPI)  # Group nodes by their degree
    
    for i in range(num_iterations):  # for each iteration
        iteration_results = {}
        for j, drug in enumerate(df["Drugs"]):  # for each drug, keep its row (j)
            target_list = df["Targets"][j]  # get the list of targets for that drug
            target_list_PPI = set(target_list) & set(PPI.nodes())  # get the drug's targets that are in the PPI
            drug_targets = []
            for target in target_list_PPI:  # for each target in the target list
                degree_target = PPI.degree(target)  # calculate its degree
                available_targets = group_nodes_by_degree[degree_target]  # choose proteins from the PPI with the same degree
                random_target = np.random.choice(available_targets)  # same number of nodes as the disease module, taken from the total list of PPI proteins randomly
                drug_targets.append(random_target)
            iteration_results[drug] = drug_targets
        results[i] = iteration_results

    df_results = pd.DataFrame(results)
    return df_results

# ================================================================================= 

def determine_treatment(row, dis_dru_the):
    disease = row['ID']
    drug = row['Drugs']

    treatments = dis_dru_the[(dis_dru_the['dis'] == disease) & (dis_dru_the['dru'] == drug)]

    treatment = 'yes' if treatments.shape[0] > 0 else 'unknown'

    return treatment

# =================================================================================

def rep_prox(df_combined, disease_name):
    """
    This function represents in a boxplot the distribution of proximity to a disease and
    the distribution of its z-score for the group of drugs used to treat the disease
    and the group of drugs not used for its treatment (unknown).

    Input:
    1. df_combined: DataFrame combined with drugs, their observed proximity, their average random proximity,
    the standard deviation of random proximity, their z-score, a column indicating the disease
    (dementia, epilepsy, bipolar disorder, or schizophrenia), and another column indicating if the drug
    is used for the treatment of the disease in the same row.
    2. disease_name: Name of the disease
    """
    # Filter the data to generate two new dataframes, one with data for drugs used as treatment for the disease, and another with data for the rest of drugs
    drugs_with_disease = df_combined[(df_combined['ID'] == disease_name) & (df_combined['Treatment'] == 'yes')]
    drugs_without_disease = df_combined[(df_combined['ID'] == disease_name) & (df_combined['Treatment'] == 'unknown')]

    combined_data = pd.concat([drugs_with_disease.assign(Treatment='Treatment'), drugs_without_disease.assign(Treatment='Unknown')])

    # Combine the two datasets into a single subplot
    fig, axes = plt.subplots(1, 2, figsize=(14, 6))  # Create a subplot with 1 row and 2 columns

    # Plot the boxplot with both proximity distributions
    sns.boxplot(x='Treatment', y='Closest distance', data=combined_data, hue='Treatment', ax=axes[0], palette={'Treatment': '#FF7A7A', 'Unknown': '#79C4FF'}, dodge=False, medianprops=dict(linewidth=2), legend=False)
    axes[0].set_ylabel('Closest distance ($\mathregular{d_c}$)', fontsize=12)
    axes[0].set_xlabel('')
    for label in axes[0].get_xticklabels():
        label.set_fontsize(12)

    # Plot the boxplot with both proximity distributions
    sns.boxplot(x='Treatment', y='Dc_zscore', data=combined_data, hue='Treatment', ax=axes[1], palette={'Treatment': '#FF7A7A', 'Unknown': '#79C4FF'}, dodge=False, medianprops=dict(linewidth=2), legend=False)
    axes[1].set_ylabel('Proximity [z-score ($\mathregular{d_c}$)]', fontsize=12)
    axes[1].set_xlabel('')
    for label in axes[1].get_xticklabels():
        label.set_fontsize(12)

    plt.tight_layout()  # Adjust the layout of the subplot to avoid overlap
    plt.show()

# =================================================================================